
// Process Arrays

#include <stdlib.h>
//#include <iostream.h>
extern "C" {
  int printf(char*, ...) ;
}
#include <stdarg.h>
extern "C" {
  #include "mpi.h"
}
#include "Seq.h"
#include "Procs.h"
#include "Threads.h"
#include "ProcTree.h"

int mapVirtuals(int [], int [], int, int) ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Global `Adlib' operations.                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Procs* world ;
ActiveProcessGroup* apg ;


// Following two operations are globally collective

void AdlibInit(int& argc, char** (&argv)) {
  MPI_Init(&argc, &argv) ;

  world = new Procs ;
  apg = new ActiveProcessGroup(world) ;

  ths = new Threads ;
}


void AdlibFinalize() {
  delete ths ;

  delete apg ;
  delete world ;

  MPI_Finalize() ;
}
 
void AdlibPrintf(char* message, ...) {
  va_list ap ;

  {
    int locId ;
    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;

    printf("Process %d: ", locId) ;
  }

  va_start(ap, message) ;

  char* substring = message ;
  int datum ;
  for(char* ch = message ; ; ch++)
    switch(*ch) {
      case 0 :
        printf(substring, datum) ;
        return ;
      case '%' :
        {
          int sublen = ch - substring ;
          char* subcopy = new char [sublen + 1] ;

          memcpy(subcopy, substring, sublen) ;
          subcopy [sublen] = 0 ;
          printf(subcopy, datum) ;

          delete subcopy ;
        }
        substring = ch ;
        datum = va_arg(ap, int) ;  // No double-word arguments!
        break ;
      default :
        break ;
    }

  va_end(ap) ;
}


void AdlibError(char* message) {
    printf("%s\n", message) ;
  MPI_Abort(MPI_COMM_WORLD, 1) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `SubProcs': Process sub-arrays.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


SubProcs :: SubProcs() {}

SubProcs :: SubProcs(Procs* _procs) : procs(_procs),
				      rank(_procs->rnk()) {}

SubProcs :: SubProcs(const SubProcs& subProcs) : procs(subProcs.procs),
				                 rank(subProcs.rank) {}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `ActiveProcessGroup': the process group executing this     //
//                                control thread.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


ActiveProcessGroup :: ActiveProcessGroup() {}

ActiveProcessGroup :: ActiveProcessGroup(Procs* procs) :
    SubProcs(procs) {}

ActiveProcessGroup :: ActiveProcessGroup(const ActiveProcessGroup& apg) :
    SubProcs(apg) {}


int ActiveProcessGroup :: adjustProcs(Procs* _procs) {
  // Make parent process array `_procs'.

  // Unless the parent process group is already `_procs'
  // (in which case this is a null operation), this call is
  // only allowed if parent process group of `_procs' is identical
  // to the process group defined by the current state of `this'.

  if(_procs != procs) {
    if(_procs->parent.procs == procs && _procs->parent.rank == rank) {
      rank  = _procs->rnk() ;
      procs = _procs ;

      return !(procs->outsider) ;
    }
    else
      AdlibError("ActiveProcessGroup :: adjustProcs: "
                 "inaccessible process array") ;
  }
  else
    return 1 ;
}

void ActiveProcessGroup :: restoreProcs() {
  // Undo the matching `adjustProcs' operation.

  if(rank == procs->rnk()) {
    // All dimensions of `procs' reactivated.  Revert to parent process group.
    rank  = procs->parent.rank ;
    procs = procs->parent.procs ;
  }
}


void ActiveProcessGroup :: deactivate(Dimension* dim) {
  if(dim->procs != procs)
    AdlibError("ActiveProcessGroup :: deactivate: "
                   "dimension not in active process array") ;

  if(dim->active) {
    dim->active = 0 ;
    rank-- ;
  }
  else
    AdlibError("ActiveProcessGroup :: deactivate: inactive dimension array") ;
}

int ActiveProcessGroup :: deactivate(Dimension* dim, const int coord) {
  deactivate(dim) ;

  return coord == dim->coord ;
}

void ActiveProcessGroup :: reactivate(Dimension* dim) {
  rank++ ;
  dim->active = 1 ;
}


int ActiveProcessGroup :: deactivate(Coord* coord) {
  return deactivate(coord->dim, coord->coord) ;
}

void ActiveProcessGroup :: reactivate(Coord* coord) {
  reactivate(coord->dim) ;
}


int ActiveProcessGroup :: on(ProcessGroup* p) {
  int onFlag = p->procs->on() ;

  for(int i = 0 ; i < procs->rnk() ; i++) {
    Dimension* dim = procs->dim(i + 1) ;
    if(dim->active && !p->active [i]) {
      onFlag &= deactivate(dim, p->coords [i]) ;
      p->atMask [i] = 1 ;
    }
    else
      p->atMask [i] = 0 ;
  }

  return onFlag ;
}

void ActiveProcessGroup :: no(ProcessGroup* p) {
  for(int i = procs->rnk() - 1 ; i >= 0 ; i--)
    if(p->atMask [i])
      reactivate(procs->dim(i + 1)) ;

  p->procs->no() ;
}


int ActiveProcessGroup :: size() const {
  // Number of processes in the APG.

  int size = 1 ;
  for(int i = 0 ; i < procs->rnk() ; i++) {
    Dimension* dim = procs->dim(i + 1) ;
    if(dim->active)
      size *= dim->extent ;
  }

  return size ;
}

int ActiveProcessGroup :: locIndex() const {
  // Index of local process in an enumeration of the APG.

  int index = 0 ;
  for(int i = 0 ; i < procs->rnk() ; i++) {
    Dimension* dim = procs->dim(i + 1) ;
    if(dim->active)
      index = (dim->extent * index) + dim->coord ;
  }

  return index ;
}

int ActiveProcessGroup :: procId(const int _index) const {
  // Physical id of process with specified enumeration index in the APG.

  int index = _index ;
  int parentIndex = 0 ;
  for (int i = procs->rnk() - 1 ; i >= 0 ; i--) {
    Dimension* dim = procs->dim(i + 1) ;
    int coord ;
    if(dim->active) {
      int extent = dim->extent ;
      coord = index % extent ;
      index /= extent ;
    }
    else
      coord = dim->coord ;
    parentIndex += dim->stride * coord ;
  }

  return (procs->physIds) [parentIndex] ;
}


void ActiveProcessGroup :: spanTree(ProcTree* tree) {

  const int index  = locIndex() ;
  const int extent = size() ;

  tree->numChildren = 0 ;
  for(int i = 1 ; i <= tree->a ; i++) {
    const int ch = tree->a * index + i ;
    if(ch < extent) {
      tree->child [i - 1] = procId(ch) ;
      tree->numChildren++ ;
    }
  }

  if(index == 0)
    tree->isRoot = 1 ;
  else {
    tree->isRoot = 0 ;
    const int pr = (index - 1) / tree->a ;
    tree->parent = procId(pr) ;
  }
}


void ActiveProcessGroup :: radialTree(ProcTree* tree, ProcessGroup* spine) {

  // APG and `spine' must be part of the same process array, the
  // latter a subarray of the former.

  // The computed process tree connects the spine with all other
  // processes in the APG, via dimensions orthogonal to those of the
  // spine.

  int spineIndex = 0 ;
  int i = 0 ;

  while(i < procs->rnk()) {
    Dimension* dim = procs->dim(i + 1) ;
    if(spine->active [i])
      deactivate(dim) ;
    else
      if(dim->active)
	spineIndex = (dim->extent) * spineIndex + spine->coords [i] ;
      else
	if(spine->coords [i] != dim->coord)
	  AdlibError("ActiveProcessGroup :: radialTree:"
		     "Spine outside active process group") ;
    i++ ;
  }

  // ...APG is now the original APG, with the `spine' subspace
  // factored out.  That subspace intersects the current APG at
  // precisely one point.  `spineIndex' is index within the APG of
  // that process.

  const int extent      = size() ;
  const int indexOffset = (locIndex() + extent - spineIndex) % extent ;

  tree->numChildren = 0 ;
  for(int j = 1 ; j <= tree->a ; j++) {
    const int ch = tree->a * indexOffset + j ;
    if(ch < extent) {
      tree->child [j - 1] = procId((spineIndex + ch) % extent) ;
      tree->numChildren++ ;
    }
  }

  if(indexOffset == 0)
    tree->isRoot = 1 ;
  else {
    tree->isRoot = 0 ;
    const int pr = (indexOffset - 1) / tree->a ;
    tree->parent = procId((spineIndex + pr) % extent) ;
  }

  // Restore APG to original state...

  while(i > 0) {
    i-- ;
    Dimension* dim = procs->dim(i + 1) ;
    if(spine->active [i])
      reactivate(dim) ;
  }
}


// Small auxilliary function---should be inline

void ActiveProcessGroup :: decrement(int* axis) {
  while((--*axis) >= 0 && ! procs->dim(*axis + 1)->active) ;
      // Find previous active dimension
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `ProcessGroup'.                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


ProcessGroup :: ProcessGroup(ProcessGroup& procGrp) : SubProcs(procGrp) {

  // Copy constructor.

  active = new int [procs->rnk()] ;
  coords = new int [procs->rnk()] ;
  for(int i = 0 ; i < procs->rnk() ; i++) {
    active [i] = procGrp.active [i] ;
    coords [i] = procGrp.coords [i] ;
  }
  lead = procGrp.lead ;

  lastAxis = procGrp.lastAxis ;

  atMask = new int [procs->rnk()] ;
}

ProcessGroup :: ProcessGroup(Procs& _procs) : SubProcs(&_procs) {

  // Conversion constructor; construct `ProcessGroup' describing
  // given process array.

  // Check `parent' of `procs' is current active process group
  if(procs->parent.procs != apg->procs || procs->parent.rank != apg->rank)
    AdlibError(
        "ProcessGroup :: ProcessGroup(Procs&): inaccessible process array") ;

  active = new int [procs->rnk()] ;
  coords = new int [procs->rnk()] ;
  for(int i = 0 ; i < procs->rnk() ; i++) {
    active [i] = 1 ;
    coords [i] = 0 ;
  }
  lead = 0 ;

  lastAxis = procs->rnk() - 1 ;

  atMask = new int [procs->rnk()] ;
}

ProcessGroup :: ProcessGroup(ActiveProcessGroup& apg) : SubProcs(apg) {

  // Construct `ProcessGroup' describing active process group.

  active = new int [procs->rnk()] ;
  coords = new int [procs->rnk()] ;
  for(int i = 0 ; i < procs->rnk() ; i++) {
    Dimension* dim = procs->dim(i + 1) ;

    if(active [i] = dim->active)
      coords [i] = 0 ;
    else
      coords [i] = dim->coord ;
  }
  lead = apg.procId(0) ;

  lastAxis = procs->rnk() ;
  decrement(&lastAxis) ;

  atMask = new int [procs->rnk()] ;
}

ProcessGroup :: ~ProcessGroup() {
  delete atMask ;
  delete coords ;
  delete active ;
}


void ProcessGroup :: deactivate(const Dimension* dim, const int coord) {
  if(dim->procs != procs) {
    printf("ProcessGroup :: deactivate: dim = %d dim->procs = %x procs = %x\n",
           dim, dim->procs, procs) ;  // debug
    AdlibError("ProcessGroup :: deactivate: dimension not in process array") ;
  }

  int axis = dim->axis ;

  if(!active [axis]) {
    if(coord == coords [axis])
      return ;  // ok---treat it as a null operation
    else
      AdlibError("ProcessGroup :: deactivate: inactive dimension") ;
  }

  active [axis] = 0 ;
  rank-- ;
  coords [axis] = coord ;
  lead += coord * dim->stride ;

  if(axis == lastAxis)
    decrement(&lastAxis) ;
}

void ProcessGroup :: reactivate(const Dimension* dim) {
  int axis = dim->axis ;

  lead -= coords[axis] * dim->stride ;
  coords [axis] = 0 ;
  rank++ ;
  active[axis] = 1 ;

  if(axis > lastAxis)
    lastAxis = axis ;
}

void ProcessGroup :: deactivate(const Coord* coord) {
  deactivate(coord->dim, coord->coord) ;
}

void ProcessGroup :: reactivate(const Coord* coord) {
  reactivate(coord->dim) ;
}


int ProcessGroup :: isActive(const Dimension* dim) const {
  if(dim->procs != procs)
    AdlibError("ProcessGroup :: isActive: dimension not in process array") ;

  return active[dim->axis] ;
}


int ProcessGroup :: procId() const {
  // Physical id of lead process of group

  return procs->physIds [lead] ;
}


// Iterator functions: enumerate all processes in the group.
// (The order of enumeration has last active dimension most rapidly varying.)

void ProcessGroup :: enumerate() {
  testFlag = 1 ;
}

int  ProcessGroup :: test() {return testFlag ; }

void ProcessGroup :: next() {
  int axis = lastAxis ;

  loop :
    if(axis >= 0) {
      const Dimension* dim = procs->dim(axis + 1) ;
      int* coord = coords + axis ;

      (*coord)++ ;
      lead += dim->stride ;
      if(*coord == dim->extent) {
        *coord = 0 ;
        lead -= dim->stride * dim->extent ;
        decrement(&axis) ;
        goto loop ;
      }
    }
    else
      testFlag = 0 ;
}


// Support for translation of `on' statement.

//int ProcessGroup :: on() {
//
//  int onFlag = procs->on() ;
//
//  for(int i = 0 ; i < procs->rnk() ; i++) {
//    Dimension* dim = procs->dim(i + 1) ;
//    if(dim->active && !active [i]) {
//      onFlag &= apg->deactivate(dim, coords [i]) ;
//      atMask [i] = 1 ;
//    }
//    else
//      atMask [i] = 0 ;
//  }
//
//  return onFlag ;
//}
//
//void ProcessGroup :: no() {
//
//  for(int i = procs->rnk() - 1 ; i >= 0 ; i--)
//    if(atMask [i])
//      apg->reactivate(procs->dim(i + 1)) ;
//
//  procs->no() ;
//}

int ProcessGroup :: on() {
  return apg->on(this) ;
}

void ProcessGroup :: no() {
  apg->no(this) ;
}


// Replace process group by its intersection with the active process group
// (support for construction of array sections).

void ProcessGroup :: activeSection() {
  
  if(procs == apg->procs)
    for(int i = procs->rnk() - 1 ; i >= 0 ; i--) {
      Dimension* dim = procs->dim(i + 1) ;
      //if(!dim->active) {
      //  deactivate(dim, dim->coords) ;  // doesn't work???
      if(!dim->active) {
	if(!active [i]) {
	  if(dim->coord == coords [i])
	    return ;  // ok---treat it as a null operation
	  else
	    AdlibError("ProcessGroup :: activeSection: Section is empty") ;
	}

	active [i] = 0 ;
	rank-- ;
	coords [i] = dim->coord ;
	lead += dim->coord * dim->stride ;

	if(i == lastAxis)
	  decrement(&lastAxis) ;
      }
  }
  else {
    // assert(apg == procs.parent)
    if(apg->procs != procs->parent.procs || apg->rank != procs->parent.rank)
      AdlibError("ProcessGroup :: activeSection: Section is undefined") ;
  }
}


// Small auxilliary function---should be inline

void ProcessGroup :: decrement(int* axis) {
  while((--*axis) >= 0 && !(active [*axis])) ;
      // Find previous active dimension
}


ProcessGroup ProcessGroup :: operator/(Coord& coord) const {
  ProcessGroup result(*this) ;
  result.deactivate(&coord) ;
  return result ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Procs': the process array.                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Procs :: Procs() : rank(1), outsider(0) {            // [rank outsider :-)]

  // Default constructor for `Procs'.  Create the world---a
  // 1-dimensional process array containing the MPI physical processes.

  // Normally this constructor is invoked as part of the `Adlib'
  // startup sequence---not in user code.

  // `parent' left undefined.

  int coord, size ;
  MPI_Comm_rank(MPI_COMM_WORLD, &coord) ;
  MPI_Comm_size(MPI_COMM_WORLD, &size) ;

  physIds = new int [size] ;
  for(int i = 0 ; i < size ; i++)
    physIds [i] = i ;

  dims  = new Dimension* ;
  *dims = new Dimension(size, size, coord, 1, this, 0) ;
}


Procs :: Procs(const int _rank, int* _n) : rank(_rank), parent(*apg) {

  // The normal `Procs' constructor

  int i ;

  int* n      = new int [rank] ;
  for (i = 0; i < rank ; i++)
    n [i] = _n [i] ;

  // Initialise members...

  // Choose mapping of virtual processes into physical processes
  int* q = new int [rank] ;
  int size = mapVirtuals(q, n, rank, apg->size()) ;

  //// debug
  //if(apg.locIndex() == 0) {
  //  printf("%s", "Procs: virtual process array shape = (") ;
  //  if(rank > 0) {
  //    printf("%d", n [0]) ;
  //    for(int i = 1 ; i < rank ; i++)
  //      printf(", %d", n [i]) ;
  //  }
  //  printf("%s", ")\n") ;
  //  printf("%s", "Procs: physical process array shape = (") ;
  //  if(rank > 0) {
  //    printf("%d", q [0]) ;
  //    for(int i = 1 ; i < rank ; i++)
  //      printf(", %d", q [i]) ;
  //  }
  //  printf("%s", ")\n") ;
  //}


  int myIndex = apg->locIndex() ;

  physIds = new int [size] ;
  for(i = 0 ; i < size ; i++)
    physIds [i] = apg->procId(i) ;

  // Set `outsider' flag if this process doesn't hold any virtual processes
  outsider = myIndex >= size ;

  dims = new Dimension* [rank] ;
  int stride = 1 ;
  for (i = rank - 1; i >= 0 ; i--) {
    int coord = (myIndex / stride) % q [i] ;
    dims [i] = new Dimension(n [i], q [i], coord, stride, this, i) ;
    stride *= q [i] ;
  }

  delete q ;
  delete n ;

  onStk = 0 ;
}


Procs :: ~Procs() {
  delete physIds ;
  for(int i = 0 ; i < rank ; i++)
    delete dims [i] ;
  delete dims ;
}

// these should be inline...

int Procs :: rnk() const {return rank ; }

Dimension* Procs :: dim(const int axis) const {return dims [axis - 1] ; }

int Procs :: on() {

  // The `stack' count here is a quick fix.  Probably I have to to
  // some rationalisation.

  if(apg->procs != this)
    apg->adjustProcs(this) ;
  else
    onStk++ ;

  return !outsider ;
}

void Procs :: no() {
  if(onStk == 0)
    apg->restoreProcs() ;
  else
    onStk-- ;
}


ProcessGroup Procs :: operator/(Coord& coord) const {
  ProcessGroup result(*this) ;
  result.deactivate(&coord) ;
  return result ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Dimension': the process array dimension.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Dimension :: Dimension(const int _virExtent,
	       const int _extent, const int _coord, const int _stride,
	       Procs* _procs, const int _axis) :
    extent(_extent), coord(_coord), stride(_stride),
    virExtent(_virExtent),
    blockSize((_virExtent + _extent - 1) / _extent),
    procs(_procs),
    axis(_axis),
    active(1) {}

// ---------------------------------------------------------------------
//
// Things to do...
//
//    Should revert to an earlier implementation of `ProcessGroup :: on',
//    which used an `atStack' rather than `atMask'.
//    ***I SUSPECT THE CURRENT IMPLEMENTATION IS ERRONEOUS.***
//
//    Move the core functionality of `ProcessGroup :: on/no' into
//    `ActiveProcessGroup'?
//
//    Functionality of  `ProcessGroup :: activeSection' should be
//    subsumed into constuctor?  Dunno.  Think about it.
//
//    Find out why calling `ProcessGroup :: deactivate' from
//    `ProcessGroup :: activeSection' doesn't work.
